"""
       R. S. Sutton and A. G. Barto (2018)
       Reinforcement Learnig, 2nd ed., 25-32.
"""
import numpy as np
import matplotlib.pyplot as plt

rng = np.random.default_rng(123)    #  シミュレーションの再現性のために値を指定して初期化

k = 10
RsDic = {}
Rsave = {}
n_simu = 1000
ep_lst = [0.0, 0.01, 0.1]

n_mcs = 1000
for ep in ep_lst:
    RsDic[ep] = []
    for i_mcs in range(n_mcs):
        if i_mcs % 100 == 0:
            print(f'{ep}/ep  {i_mcs}/{n_mcs}')
        Qt = np.zeros(k)      #  Mean reward actually received
        Nt = np.zeros(k)
        q_star = rng.normal(0, 1 , k)
        Rt = []

        for i in range(n_simu):
            if rng.uniform(0.0, 1.0) < ep:
                a = rng.choice(np.arange(k))   # Exploration
            else:
                a = np.argmax(Qt)              # Greedy

            Nt[a] += 1
            R = rng.normal(q_star[a], 1, 1)[0]
            Qt[a] += (R - Qt[a]) / Nt[a]       #  平均rewardの更新
            Rt.append(R)
        RsDic[ep].append(Rt)
ep_c = {0.0:'g', 0.01:'r', 0.1:'b'}
for ep in ep_lst:
    Rsave[ep] = np.mean(RsDic[ep], axis=0)

for ep in ep_lst:
    plt.plot(np.arange(n_simu), Rsave[ep], c=ep_c[ep], lw=2,
             label=fr'$\epsilon$={ep}')
plt.xlabel('Steps')
plt.ylabel('Average reward')
plt.legend()
plt.show()


